# Copyright (C) 2004-2005 struktur AG and Contributors <simon@struktur.de>
# http://www.strukturag.com, http://www.icoya.com

# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL). A copy of the ZPL should accompany this distribution.

# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.

# See LICENSE.txt for details

""" Session Crumbler: enable session for non session user folders.

$Id: SessionCrumbler.py 36974 2007-02-09 05:22:07Z magnon $
"""

__version__ = "$Rev: 36974 $"

import Globals, sys

from base64 import encodestring, decodestring
from urllib import quote, unquote

from zLOG import LOG, ERROR
from DateTime import DateTime
from AccessControl import ClassSecurityInfo
from ZPublisher.HTTPRequest import HTTPRequest

from Products.CMFCore.CookieCrumbler import CookieCrumbler, ResponseCleanup

from SessionSecureContainer import SessionSecureContainer
from sessionlazycheck import sessionlazycheck

try:
    from zExceptions import Redirect
except ImportError:
    # Pre Zope 2.7.
    Redirect = 'Redirect'

# Constants.
ATTEMPT_NONE = 0       # No attempt at authentication
ATTEMPT_LOGIN = 1      # Attempt to log in
ATTEMPT_CONT = 2       # Attempt to resume session

class SessionCrumblerDisabled(Exception):
    """SessionCrumbler should not be used for a request."""

class SessionCrumbler(CookieCrumbler):
    '''
    Reads session during traversal and simulates the HTTP auth header.
    '''

    meta_type = 'Session Crumbler'

    name_cookie = '__ac_name'
    pw_cookie = '__ac_password'
    persist_cookie = '__ac_persistent'
    auto_login_page = 'require_login'
    unauth_page = 'require_login'
    logout_page = 'logged_out'
    cache_header_value = 'no-cache'

    security = ClassSecurityInfo()

    _properties = ({'id':'name_cookie', 'type': 'string', 'mode':'w',
                    'label':'User name form variable'},
                   {'id':'pw_cookie', 'type': 'string', 'mode':'w',
                    'label':'User password form variable'},
                   {'id':'persist_cookie', 'type': 'string', 'mode':'w',
                    'label':'User name persistence form variable'},
                   {'id':'auto_login_page', 'type': 'string', 'mode':'w',
                    'label':'Auto-login page ID'},
                   {'id':'unauth_page', 'type': 'string', 'mode':'w',
                    'label':'Failed authorization page ID'},
                   {'id':'logout_page', 'type': 'string', 'mode':'w',
                    'label':'Logout page ID'},
                   {'id':'cache_header_value', 'type': 'string', 'mode':'w',
                    'label':'Cache-Control header value'},
                   )

    security.declarePrivate('getAuthSession')
    def getAuthSession(self, session, key):
        # Get the auth from session.
        
        # Get secure container.
        try: sec = session[key]
        except: sec = None
            
        # Get credentiats from container.
        if isinstance(sec, SessionSecureContainer):
            return sec.get('ac')
            
        return None

    security.declarePrivate('setAuthSession')
    def setAuthSession(self, session, key, value):
        # Set auth into session.
        
        # Get or create secure session container.
        try: sec = session[key]
        except: sec = SessionSecureContainer()
        
        # Store credentials to secure container.
        sec.set('ac', value)
        
        # Store container into session.
        session[key]=sec

    security.declarePrivate('modifyRequest')
    def modifyRequest(self, req, resp):
        '''
        Copies session-supplied credentials to the basic auth fields.
        '''
        
        if (req.__class__ is not HTTPRequest
            or not req['REQUEST_METHOD'] in ('HEAD', 'GET', 'PUT', 'POST')
            or req.environ.has_key('WEBDAV_SOURCE_PORT')):
            raise SessionCrumblerDisabled
              
        # attempt may contain information about an earlier attempt to
        # authenticate using a higher-up cookie crumbler within the
        # same request.
        attempt = getattr(req, '_session_auth', ATTEMPT_NONE)

        if attempt == ATTEMPT_NONE:
            
            if req._auth:
                # An auth header was provided and no session crumbler
                # created it. The user must be using basic auth.
                raise SessionCrumblerDisabled
    
            if req.has_key(self.pw_cookie) and req.has_key(self.name_cookie):
                # Attempt to log in and set session authentication.
                
                attempt = ATTEMPT_LOGIN
                
                name = req[self.name_cookie]
                pw = req[self.pw_cookie]
                ac = encodestring('%s:%s' % (name, pw))
                req._auth = 'basic %s' % ac
                resp._auth = 1
                if req.get(self.persist_cookie, 0):
                    # Persist the user name (but not the pw or session).
                    expires = (DateTime() + 365).toZone('GMT').rfc822()
                    resp.setCookie(self.name_cookie, name, path='/',
                                   expires=expires)
                else:
                    # Expire the user name.
                    resp.expireCookie(self.name_cookie, path='/')
                    
                # Set authentication to session.
                self.setAuthSession(req.SESSION, self.auth_cookie, ac)
                self.delRequestVar(req, self.name_cookie)
                self.delRequestVar(req, self.pw_cookie)
                
            elif self.requestHasSession(req):
                # Attempt to resume an existing authentication session.
                # Copy __ac to the auth header.
                ac = self.getAuthSession(req.SESSION, self.auth_cookie)
                if ac:
                    try:
                        decodestring(ac)
                    except:
                        # Not a valid auth header.
                        pass
                    else:
                        attempt = ATTEMPT_CONT
                        req._auth = 'basic %s' % ac
                        resp._auth = 1
            
        # Mark request as session authenticated.
        req._session_auth = attempt

        return attempt

    def __call__(self, container, req):
        '''
        The __before_publishing_traverse__ hook.
        '''
        
        resp = self.REQUEST['RESPONSE']
        try:
            attempt = self.modifyRequest(req, resp)
        except SessionCrumblerDisabled:
            return
        if req.get('disable_session_login__', 0):
            return

        if (self.unauth_page or
            attempt == ATTEMPT_LOGIN or attempt == ATTEMPT_NONE):
            # Modify the "unauthorized" response.
            req._hold(ResponseCleanup(resp))
            resp.unauthorized = self.unauthorized
            resp._unauthorized = self._unauthorized
        if attempt != ATTEMPT_NONE:
            # Trying to log in or resume a session.
            if self.cache_header_value:
                # we don't want caches to cache the resulting page.
                resp.setHeader('Cache-Control', self.cache_header_value)
                # demystify this in the response.
                resp.setHeader('X-Cache-Control-Hdr-Modified-By',
                               'SessionCrumbler')
            phys_path = self.getPhysicalPath()
            if self.logout_page:
                # Session login is in use.
                page = getattr(container, self.logout_page, None)
                if page is not None:
                    # Provide a logout page.
                    req._logout_path = phys_path + ('logout',)
            req._credentials_changed_path = (
                phys_path + ('credentialsChanged',))

    security.declarePublic('credentialsChanged')
    def credentialsChanged(self, user, name, pw):
        req = self.REQUEST
        ac = encodestring('%s:%s' % (name, pw))
        self.setAuthSession(req.SESSION, self.auth_cookie, ac)

    security.declarePublic('requestHasSession')
    def requestHasSession(self, req):
        # Check for a session without creating a new one.
        # (returns True or False)
        return sessionlazycheck(self)

    security.declarePrivate('unauthorized')
    def unauthorized(self):
        req = self.REQUEST
        resp = self._cleanupResponse()
        if self.requestHasSession(req):
            # If we set the auth to session before, delete it now.
            session = req['SESSION']
            if session.has_key(self.auth_cookie):
                del session[self.auth_cookie]
        # Redirect if desired.
        url = self.getUnauthorizedURL()
        if url is not None:
            raise Redirect, url
        # Fall through to the standard unauthorized() call.
        resp.unauthorized()

    def _unauthorized(self):
        req = self.REQUEST
        resp = self._cleanupResponse()
        if self.requestHasSession(req):
            # If we set the auth to session before, delete it now.
            session = req['SESSION']
            if session.has_key(self.auth_cookie):
                del session[self.auth_cookie]   
        # Redirect if desired.
        url = self.getUnauthorizedURL()
        if url is not None:
            resp.redirect(url, lock=1)
            # We don't need to raise an exception.
            return
        # Fall through to the standard _unauthorized() call.
        resp._unauthorized()

    security.declarePublic('getUnauthorizedURL')
    def getUnauthorizedURL(self):
        '''
        Redirects to the login page.
        '''
        req = self.REQUEST
        resp = req['RESPONSE']
        attempt = getattr(req, '_session_auth', ATTEMPT_NONE)
        if attempt == ATTEMPT_NONE:
            # An anonymous user was denied access to something.
            page_id = self.auto_login_page
            retry = ''
        elif attempt == ATTEMPT_LOGIN:
            # The login attempt failed. Try again.
            page_id = self.auto_login_page
            retry = '1'
        else:
            # An authenticated user was denied access to something.
            page_id = self.unauth_page
            retry = ''
        if page_id:
            page = self.restrictedTraverse(page_id, None)
            if page is not None:
                came_from = req.get('came_from', None)
                if came_from is None:
                    came_from = req.get('URL', '')
                    query = req.get('QUERY_STRING')
                    if query:
                        # Include the query string in came_from.
                        if not query.startswith('?'):
                            query = '?' + query
                        came_from = came_from + query
                url = '%s?came_from=%s&retry=%s&disable_session_login__=1' % (
                    page.absolute_url(), quote(came_from), retry)
                return url
        return None

manage_addSCForm = Globals.HTMLFile('dtml/addSC', globals())
manage_addSCForm.__name__ = 'addSC'

def manage_addSC(self, id, REQUEST=None):
    '''
    Helper to add the session crumbler.
    '''
    ob = SessionCrumbler()
    ob.id = id
    self._setObject(id, ob)
    if REQUEST is not None:
        return self.manage_main(self, REQUEST)

